-- |
-- Module:      Math.NumberTheory.ArithmeticFunctions
-- Copyright:   (c) 2016 Andrew Lelechenko
-- Licence:     MIT
-- Maintainer:  Andrew Lelechenko <andrew.lelechenko@gmail.com>
--
-- This module provides an interface for defining and manipulating
-- arithmetic functions. It also defines several most widespreaded
-- arithmetic functions.
--

module Math.NumberTheory.ArithmeticFunctions
  ( module Math.NumberTheory.ArithmeticFunctions.Class
  , module Math.NumberTheory.ArithmeticFunctions.Standard
  ) where

import Math.NumberTheory.ArithmeticFunctions.Class
import Math.NumberTheory.ArithmeticFunctions.Standard
